# coding=utf-8
import copy
import BiliUtil.Util as Util


class Video:
    def __init__(self, aid, cid, name, page):
        self.aid = str(aid)
        self.cid = str(cid)
        self.name = name
        self.page = page
        self.quality = None
        self.length = None
        self.format = None
        self.height = None
        self.width = None
        self.level = None
        self.video = None
        self.audio = None

    def sync(self, cookie=None, quality=None):
        # 检验必要的参数
        if self.aid is None or self.cid is None:
            raise Util.ParameterError('缺少获取视频信息的必要参数')

        if quality is None:
            quality = Util.Config.Quality.V1080P60

        # 发送网络请求
        http_request = {
            'info_obj': Util.VIDEO,
            'params': {
                'avid': str(self.aid),
                'cid': str(self.cid),
                'qn': quality[0],
                'otype': 'json',
                'fnver': 0,
                'fnval': 16
            },
            'cookie': cookie
        }
        json_data = Util.http_get(**http_request)

        # 自动识别不同的数据
        self.format = json_data['data']['format']
        self.length = json_data['data']['timelength']
        self.quality = Util.Config.Quality.INDEX[json_data['data']['quality']]
        if 'dash' in json_data['data']:
            self.level = 'new_version'
            video_obj = json_data['data']['dash']['video'][0]
            audio_obj = json_data['data']['dash']['audio'][0]
            self.height = video_obj['height']
            self.width = video_obj['width']
            self.video = list([video_obj['baseUrl']])
            self.audio = list([audio_obj['baseUrl']])
            if video_obj['backup_url']:
                for backup in video_obj['backup_url']:
                    self.video.append(backup)
            if audio_obj['backup_url']:
                for backup in audio_obj['backup_url']:
                    self.video.append(backup)

        elif 'durl' in json_data['data']:
            self.level = 'old_version'
            video_obj = json_data['data']['durl'][-1]
            self.video = list([video_obj['url']])
            if video_obj['backup_url']:
                for backup in video_obj['backup_url']:
                    self.video.append(backup)

        # 返回视频信息
        return copy.deepcopy(vars(self))



# coding=utf-8
import re
import copy
from urllib import parse
import BiliUtil.Util as Util
import BiliUtil.Video as Video


class Album:
    def __init__(self, aid=None):
        self.aid = str(aid)
        self.num = None
        self.type = None
        self.cover = None
        self.name = None
        self.time = None
        self.desc = None
        self.view = None
        self.danmu = None
        self.reply = None
        self.favorite = None
        self.coin = None
        self.share = None
        self.like = None
        self.cid_list = None

    def set_album(self, aid):
        self.aid = str(aid)

    def set_by_url(self, url):
        input_url = parse.urlparse(url)
        aid = re.match('/video/av([0-9]+)', input_url.path).group(1)
        self.aid = str(aid)

    def sync(self, cookie=None):
        # 检验必要的参数
        if self.aid is None:
            raise Util.ParameterError('缺少获取视频信息的必要参数')

        # 发送网络请求
        http_request = {
            'info_obj': Util.ALBUM,
            'params': {
                'aid': str(self.aid)
            },
            'cookie': cookie
        }
        json_data = Util.http_get(**http_request)

        # 修改对象信息
        self.num = json_data['data']['videos']
        self.type = json_data['data']['tname']
        self.cover = json_data['data']['pic']
        self.name = json_data['data']['title']
        self.time = json_data['data']['ctime']
        self.desc = json_data['data']['desc']
        self.view = json_data['data']['stat']['view']
        self.danmu = json_data['data']['stat']['danmaku']
        self.reply = json_data['data']['stat']['reply']
        self.favorite = json_data['data']['stat']['favorite']
        self.coin = json_data['data']['stat']['coin']
        self.share = json_data['data']['stat']['share']
        self.like = json_data['data']['stat']['like']
        self.cid_list = list(page['cid'] for page in json_data['data']['pages'])

        # 返回专辑信息
        return copy.deepcopy(vars(self))

    def get_video_list(self, cookie=None):
        # 检验必要的参数
        if self.aid is None:
            raise Util.ParameterError('缺少获取视频信息的必要参数')

        if self.cid_list is None:
            self.sync(cookie)

        video_list = []
        for index, cid in enumerate(self.cid_list):
            cv = Video.Video(self.aid, cid, self.name, index+1)
            video_list.append(cv)

        return video_list



# coding=utf-8
import os
import re
import warnings

from BiliUtil.Util import Config

from BiliUtil.Space import User
from BiliUtil.Space import Channel

from BiliUtil.Video import Album
from BiliUtil.Video import Video

_aria2c_result = os.popen('aria2c -v').read()
_aria2c_result = re.match(r'(\S+)', _aria2c_result)
_ffmpeg_result = os.popen('ffmpeg -version').read()
_ffmpeg_result = re.match(r'(\S+)', _ffmpeg_result)
if _aria2c_result is not None and _aria2c_result.group(1) == 'aria2':
    if _ffmpeg_result is not None and _ffmpeg_result.group(1) == 'ffmpeg':
        from BiliUtil.Video import Task
        from BiliUtil.Video import Filter
        from BiliUtil.Video import Fetcher
    else:
        warnings.warn("QAQ~ 您未配置ffmpeg渲染环境，Task, Filter, Fetcher类不可用")
else:
    warnings.warn("QAQ~ 您未配置aria2c下载环境，Task, Filter, Fetcher类不可用")





# coding=utf-8
import re
from urllib import parse
import BiliUtil.Util as Util
import BiliUtil.Video as Video


class Channel:
    def __init__(self, uid=None, cid=None):
        self.uid = str(uid)
        self.cid = str(cid)
        self.name = None
        self.cover = None
        self.count = None

    def set_channel(self, uid, cid):
        self.uid = str(uid)
        self.cid = str(cid)

    def set_by_url(self, url):
        input_url = parse.urlparse(url)
        uid = re.match('/([0-9]+)/channel/detail', input_url.path).group(1)
        cid = parse.parse_qs(input_url.query)['cid'][0]
        self.uid = str(uid)
        self.cid = str(cid)

    def get_album_list(self, cookie=None):
        # 检验必要的参数
        if self.uid is None or self.cid is None:
            raise Util.ParameterError('缺少获取频道列表的必要参数')

        # 发送网络请求
        http_request = {
            'info_obj': Util.CHANNEL,
            'params': {
                'mid': str(self.uid),
                'cid': str(self.cid),
                'pn': 1,  # 当前页码下标
                'ps': 30,  # 每页视频数量
                'order': 0  # 默认排序
            },
            'cookie': cookie
        }
        album_list = []
        while True:
            json_data = Util.http_get(**http_request)

            # 修改对象信息
            self.name = json_data['data']['list']['name']
            self.cover = json_data['data']['list']['cover']
            self.count = str(json_data['data']['list']['count'])

            # 循环获取列表
            album_list.extend([Video.Album(av['aid']) for av in json_data['data']['list']['archives']])
            if len(album_list) < int(json_data['data']['page']['count']):
                http_request['params']['pn'] += 1
            else:
                break

        return album_list



# coding=utf-8
import os
import BiliUtil.Util as Util
import BiliUtil.Space as Space
import BiliUtil.Video as Video


class Fetcher:
    def __init__(self, obj):
        self.obj = obj
        self.info_list = None
        self.task_list = None
        self.exist_list = None

    def fetch_all(self, cookie=None, name_pattern=Util.Config.SET_AS_CODE, quality=None):
        """
        自动获取用户与频道名下的所有视频信息
        :param cookie: 权限信息
        :param quality: 下载视频质量
        :param name_pattern: 路径命名方式
        :return: 所有视频的av号列表
        """
        av_list = []
        self.info_list = []

        # 确定对象类型
        if isinstance(self.obj, Space.User):
            obj_code = self.obj.uid
        elif isinstance(self.obj, Space.Channel):
            obj_code = self.obj.cid
        else:
            raise Util.ParameterError('该类型对象无法自动加载')

        # 逐级确定储存路径
        if name_pattern == Util.Config.SET_AS_NAME:
            self.obj.sync()
            obj_name = Util.legalize_name(self.obj.name)
        else:
            obj_name = obj_code

        album_list = self.obj.get_album_list()
        for album in album_list:
            if name_pattern == Util.Config.SET_AS_NAME:
                album.sync()
                album_name = Util.legalize_name(album.name)
            else:
                album_name = album.aid

            video_list = album.get_video_list()
            for video in video_list:
                video.sync(cookie, quality)
                if name_pattern == Util.Config.SET_AS_NAME:
                    video_name = Util.legalize_name(video.name)
                else:
                    video_name = video.aid
                video_name = '{}_P{}_{}'.format(video_name, video.page, video.quality[1])

                self.info_list.append({
                    'obj': self.obj,
                    'obj_name': obj_name,
                    'album': album,
                    'album_name': album_name,
                    'video': video,
                    'video_name': video_name
                })
                av_list.append(album.aid)
        av_list = list(set(av_list))
        return av_list

    def load_task(self, output, exclude=None, v_filter=None):
        """
        自动完成下载任务的批量生成
        :param output: 输出路径
        :param exclude: 排除列表
        :param v_filter:
        :return: 下载列表中视频av号
        """
        task_list = []
        self.task_list = []
        base_path = os.path.abspath(output)
        exclude = [str(item) for item in exclude]

        # 逐个过滤并生成任务
        for info in self.info_list:
            # 执行过滤策略
            if info['album'].aid in exclude:
                continue
            elif v_filter is not None and v_filter.check_video(info['video']):
                continue

            # 创建新的下载任务
            full_path = '{}/{}/{}'.format(base_path, info['obj_name'], info['album_name'])
            self.task_list.append(Video.Task(info['video'], full_path,
                                             info['video_name'], info['album'].cover))
            task_list.append(info['album'].aid)

        task_list = list(set(task_list))
        return task_list

    def load_exist(self, output):
        """
        输出缓存中已经存在的视频编号
        区分乐观策略与悲观策略
        :param output: 输出路径
        :return: 乐观列表， 悲观列表
        """
        all_video_list = []
        positive_list = []
        negative_list = []
        base_path = os.path.abspath(output)

        for info in self.info_list:
            all_video_list.append(info['album'].aid)
            video_path = '{}/{}/{}/{}.mp4'.format(base_path, info['obj_name'], info['album_name'], info['video_name'])
            if os.path.exists(video_path):
                positive_list.append(info['album'].aid)
            else:
                negative_list.append(info['album'].aid)

        all_video_list = set(all_video_list)
        positive_list = list(set(positive_list))
        negative_list = list(all_video_list - set(negative_list))

        """
        生成策略在于
        乐观列表：Album下存在视频即认为存在
        悲观列表：Album下所有视频存在才存在
        """

        return positive_list, negative_list

    def pull_all(self, show_process=True, no_repeat=True):
        """
        批量开始下载任务
        :param show_process: 是否显示下载进度信息
        :param no_repeat: 是否重复下载
        :return: 完成下载的av号列表
        """
        av_list = []
        for task in self.task_list:
            av_list.append(task.start(show_process, no_repeat))
        av_list = list(set(av_list))
        return av_list



# coding=utf-8
from .tools import Config

from .tools import HEADER
from .tools import USER
from .tools import USER_VIDEO
from .tools import ALBUM
from .tools import VIDEO
from .tools import CHANNEL
from .tools import CHANNEL_LIST
from .tools import http_get
from .tools import legalize_name
from .tools import aria2c_pull
from .tools import ffmpeg_merge
from .tools import ParameterError
from .tools import RunningError



# coding=utf-8
import re
import copy
from urllib import parse
import BiliUtil.Util as Util
import BiliUtil.Space as Space
import BiliUtil.Video as Video


class User:
    def __init__(self, uid=None):
        self.uid = str(uid)
        self.name = None
        self.birthday = None
        self.title = None
        self.face = None
        self.time = None
        self.level = None
        self.sex = None
        self.sign = None
        self.vip = None

    def set_user(self, uid):
        self.uid = str(uid)

    def set_by_url(self, url):
        input_url = parse.urlparse(url)
        uid = re.match('/([0-9]+)', input_url.path).group(1)
        self.uid = str(uid)

    def sync(self, cookie=None):
        # 检验必要的参数
        if self.uid is None:
            raise Util.ParameterError('缺少获取用户信息的必要参数')

        # 发送网络请求
        http_request = {
            'info_obj': Util.USER,
            'params': {
                'mid': str(self.uid),
                'jsonp': 'jsonp'
            },
            'cookie': cookie
        }
        json_data = Util.http_get(**http_request)

        # 修改对象信息
        self.name = json_data['data']['name']
        self.sex = json_data['data']['sex']
        self.face = json_data['data']['face']
        self.sign = json_data['data']['sign']
        self.level = json_data['data']['level']
        self.birthday = json_data['data']['birthday']
        self.title = json_data['data']['official']['title']
        self.vip = bool(json_data['data']['vip']['status'])

        # 返回用户信息
        return copy.deepcopy(vars(self))

    def get_channel_list(self, cookie=None):
        if self.uid is None:
            raise BaseException('缺少获取频道列表的必要参数')

        # 发送网络请求
        http_request = {
            'info_obj': Util.CHANNEL_LIST,
            'params': {
                'mid': str(self.uid),
                'guest': False,
                'jsonp': 'jsonp'
            },
            'cookie': cookie
        }
        json_data = Util.http_get(**http_request)
        channel_list = list(Space.Channel(self.uid, ch['cid']) for ch in json_data['data']['list'])

        # 返回频道列表
        return channel_list

    def get_album_list(self, cookie=None):
        # 检验必要的参数
        if self.uid is None:
            raise Util.ParameterError('缺少获取视频列表的必要参数')

        # 发送网络请求
        http_request = {
            'info_obj': Util.USER_VIDEO,
            'params': {
                'mid': str(self.uid),
                'pagesize': 30,
                'tid': 0,
                'page': 1,
                'order': 'pubdate'
            },
            'cookie': cookie
        }
        album_list = []
        while True:
            json_data = Util.http_get(**http_request)

            # 循环获取列表
            album_list.extend([Video.Album(av['aid']) for av in json_data['data']['vlist']])
            if len(album_list) < int(json_data['data']['count']):
                http_request['params']['page'] += 1
            else:
                break

        # 返回视频列表
        return album_list



# coding=utf-8
import os
import re
import time
import json
import requests
import subprocess
from urllib import parse
from fake_useragent import UserAgent


# from BiliUtil import http_proxy, https_proxy


class Config:
    SET_AS_NAME = 1
    SET_AS_CODE = 2

    HTTP_PROXY = None  # http://user:pass@1.2.3.4:5678
    HTTPS_PROXY = None  # https://user:pass@1.2.3.4:5678

    class Quality:
        V360P = ('16', '流畅 360P')
        V480P = ('32', '清晰 480P')
        V720P = ('64', '高清 720P')
        V720P60 = ('74', '高清 720P60')
        V1080P = ('80', '高清 1080P')
        V1080Px = ('112', '高清 1080P+')
        V1080P60 = ('116', '高清 1080P60')
        INDEX = {
            16: V360P,
            32: V480P,
            64: V720P,
            74: V720P60,
            80: V1080P,
            112: V1080Px,
            116: V1080P60
        }


# ##################  Private  ###################

HEADER = {
    'Host': '',
    'Connection': 'keep-alive',
    'Accept': 'application/json, text/plain, */*',
    'User-Agent': '',
    'Origin': 'https://www.bilibili.com',
    'Referer': '',
    'Accept-Encoding': 'gzip, deflate, br',
    'Accept-Language': 'zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7'
}
USER = {
    'url': 'http://api.bilibili.com/x/space/acc/info',
    'origin': 'https://space.bilibili.com',
    'referer': 'https://space.bilibili.com'
}
USER_VIDEO = {
    'url': 'http://space.bilibili.com/ajax/member/getSubmitVideos',
    'origin': 'https://space.bilibili.com',
    'referer': 'https://space.bilibili.com'
}
ALBUM = {
    'url': 'http://api.bilibili.com/x/web-interface/view',
    'origin': 'https://www.bilibili.com',
    'referer': 'https://www.bilibili.com/video'
}
VIDEO = {
    'url': 'http://api.bilibili.com/x/player/playurl',
    'origin': 'https://www.bilibili.com',
    'referer': 'https://www.bilibili.com/video'
}
CHANNEL = {
    'url': 'http://api.bilibili.com/x/space/channel/video',
    'origin': 'https://space.bilibili.com',
    'referer': 'https://space.bilibili.com'
}
CHANNEL_LIST = {
    'url': 'http://api.bilibili.com/x/space/channel/list',
    'origin': 'https://space.bilibili.com',
    'referer': 'https://space.bilibili.com'
}


def http_header(info_obj):
    header = HEADER.copy()
    useragent_path = os.path.dirname(__file__) + '/fake_useragent.json'
    header['Host'] = parse.urlparse(info_obj['url']).netloc
    header['User-Agent'] = UserAgent(path=useragent_path).random
    header['Origin'] = info_obj['origin']
    header['Referer'] = info_obj['referer']
    return header


def http_get(info_obj, params, cookie=None):
    # 获取代理配置
    proxies = {
        'http': Config.HTTP_PROXY,
        'https': Config.HTTPS_PROXY
    }

    # 获取请求头信息
    new_http_header = http_header(info_obj)
    if cookie is not None:
        if isinstance(cookie, dict):
            cookie = {
                'SESSDATA': cookie['SESSDATA']
            }
        elif isinstance(cookie, str) and len(cookie) > 0:
            for line in cookie.split(';'):
                name, value = line.strip().split('=', 1)
                if name == 'SESSDATA':
                    cookie = {
                        'SESSDATA': value
                    }
                    break
        else:
            cookie = dict()
    else:
        cookie = dict()

    # 尝试进行网络连接
    for times in range(5):
        try:
            http_result = requests.get(url=info_obj['url'], params=params, cookies=cookie,
                                       headers=new_http_header, timeout=5, proxies=proxies)
        except requests.exceptions:
            time.sleep(3)
            continue
        # 尝试理解并验证响应数据
        if http_result.status_code == 200:
            try:
                json_data = json.loads(http_result.text)
            except json.decoder.JSONDecodeError:
                time.sleep(1)
                continue
            if 'code' in json_data and json_data['code'] == 0:
                return json_data
            elif 'status' in json_data and json_data['status'] is True:
                return json_data
            elif json_data['code'] == -404 or json_data['code'] == -403:
                raise RunningError('请求对象异常或被锁定，无法获取')

    raise RunningError('多次网络请求均未能获得数据：{}'.format(info_obj['url']))


def legalize_name(name):
    legal_name = re.sub(r"[\/\\\:\*\?\"\<\>\|\s']", '_', name)
    legal_name = re.sub(r'[‘’]', '_', legal_name)
    if len(legal_name) == 0:
        return 'null'
    return legal_name


def aria2c_pull(aid, path, name, url_list, show_process=False):
    # 设置输出信息
    if show_process:
        out_pipe = None
    else:
        out_pipe = subprocess.PIPE
    # 读取代理信息
    proxies = ''
    proxies += ' --http-proxy="{}"'.format(Config.HTTP_PROXY) if Config.HTTP_PROXY is not None else ""
    proxies += ' --https-proxy="{}"'.format(Config.HTTPS_PROXY) if Config.HTTPS_PROXY is not None else ""

    referer = 'https://www.bilibili.com/video/av' + str(aid)
    url = '"{}"'.format('" "'.join(url_list))
    shell = 'aria2c -c -k 1M -x {} -d "{}" -o "{}" --referer="{}" {} {}'
    shell = shell.format(len(url_list), path, name, referer, proxies, url)
    process = subprocess.Popen(shell, stdout=out_pipe, stderr=out_pipe, shell=True)
    process.wait()


def ffmpeg_merge(path, name, show_process=False):
    if show_process:
        out_pipe = None
    else:
        out_pipe = subprocess.PIPE
    flv_file = os.path.abspath('{}/{}.flv'.format(path, name))
    aac_file = os.path.abspath('{}/{}.aac'.format(path, name))
    mp4_file = os.path.abspath('{}/{}.mp4'.format(path, name))
    if os.path.exists(flv_file) and os.path.exists(aac_file):
        shell = 'ffmpeg -i "{}" -i "{}" -c copy -f mp4 -y "{}"'
        shell = shell.format(flv_file, aac_file, mp4_file)
        process = subprocess.Popen(shell, stdout=out_pipe, stderr=out_pipe, shell=True)
        process.wait()
        os.remove(flv_file)
        os.remove(aac_file)
    else:
        raise RunningError('找不到下载的音视频文件')


class ParameterError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class RunningError(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



# coding=utf-8
import os
import sys
import copy
import requests
import BiliUtil.Util as Util


class Task:
    def __init__(self, video, output, name, cover=None):
        self.video_info = copy.deepcopy(vars(video))
        del self.video_info['video']
        del self.video_info['audio']
        self.aid = video.aid
        self.level = video.level
        self.video = video.video
        self.audio = video.audio
        self.cover = cover
        self.name = name
        self.path = os.path.abspath(output)

    def start(self, show_process=True, no_repeat=True):
        """
        开始下载任务
        :param show_process: 是否显示下载进度信息
        :param no_repeat: 是否重复下载
        :return: 本次下载的视频所属的av号 / None
        """
        if not os.path.exists(self.path):
            os.makedirs(self.path)  # 创建基础路径
        # 保存视频封面
        if self.cover is not None and not os.path.exists(self.path + '/cover.jpg'):
            proxies = {
                'http': Util.Config.HTTP_PROXY,
                'https': Util.Config.HTTPS_PROXY
            }
            http_result = requests.get(self.cover, proxies=proxies)
            with open(self.path + '/cover.jpg', 'wb') as file:
                file.write(http_result.content)
        # 保存视频并转码
        if no_repeat and os.path.exists(os.path.abspath('{}/{}.mp4'.format(self.path, self.name))):
            return None
        if self.level == 'old_version':
            Util.aria2c_pull(self.aid, self.path, self.name + '.mp4', self.video, show_process)
            return self.aid
        elif self.level == 'new_version':
            Util.aria2c_pull(self.aid, self.path, self.name + '.aac', self.audio, show_process)
            Util.aria2c_pull(self.aid, self.path, self.name + '.flv', self.video, show_process)
            Util.ffmpeg_merge(self.path, self.name, show_process)
            sys.stdout.flush()
            return self.aid



# coding=utf-8
import BiliUtil.Util as Util


class Filter:
    def __init__(self, quality=None, length=None, height=None, width=None, page=None, ratio=None):
        self.quality = quality
        self.length = length
        self.height = height
        self.width = width
        self.page = page
        self.ratio = ratio

    def set_quality(self, quality):
        if ~isinstance(quality, list):
            raise Util.ParameterError('参数类型异常')
        self.quality = quality

    def set_length(self, length):
        if ~isinstance(length, list):
            raise Util.ParameterError('参数类型异常')
        self.length = length

    def set_height(self, height):
        if ~isinstance(height, list):
            raise Util.ParameterError('参数类型异常')
        self.height = height

    def set_width(self, width):
        if ~isinstance(width, list):
            raise Util.ParameterError('参数类型异常')
        self.width = width

    def set_page(self, page):
        if ~isinstance(page, list):
            raise Util.ParameterError('参数类型异常')
        self.page = page

    def set_ratio(self, ratio):
        if ~isinstance(ratio, list):
            raise Util.ParameterError('参数类型异常')
        self.ratio = ratio

    def check_video(self, video):
        """
        当响应值为True时，说明该视频超出接受范围，视频需要被过滤
        :param video: 被监测视频
        :return: 检测响应
        """
        if self.quality and video.quality not in self.quality:
            return True
        if self.length and (video.length / 1000 < self.length[0] or self.length[1] < video.length / 1000):
            return True
        if video.level == 'new_version':
            if self.height and (video.height < self.height[0] or self.height[1] < video.height):
                return True
            if self.width and (video.width < self.width[0] or self.width[1] < video.width):
                return True
            ratio = video.width / video.height
            if self.ratio and (ratio < self.ratio[0] or self.ratio[1] < ratio):
                return True
        if self.page and video.page not in self.page:
            return True
        return False
